/*
** Dialog functions
**
** Copyright  1997 Marko Mkel, see COPYING.LIB for details
*/

#include <X11/Intrinsic.h>
#include <X11/StringDefs.h>
#include <X11/Shell.h>

#include <X11/Xaw/Form.h>
#include <X11/Xaw/Label.h>
#include <X11/Xaw/Box.h>
#include <X11/Xaw/Command.h>
#include <X11/Xaw/AsciiText.h>

#include <stdio.h>
#include <stdlib.h>
#include <limits.h>

#include "util.h"
#include "dialog.h"

/* Button names and choice values */
static const struct { const char *name; Choice choice; } buttons[] = {
  { NULL, btUnspecified },
  { "yesButton", btYes },
  { "noButton", btNo },
  { "okButton", btOk },
  { "cancelButton", btCancel },
};

/* Structure for buttonCallback() to set the choice */
typedef struct {
  Choice *chosen;
  Choice choose;
} ChosenButton;

/* The button callback function */
static void buttonCallback(Widget w, ChosenButton *c, XtPointer callData);

/* Open a dialog with the specified message and buttons */
Choice Dialog (Widget parent, const char *name, unsigned whichButtons)
{
  ChosenButton choices[XtNumber(buttons)];
  Choice chosen = btUnspecified;

  Widget shell, form, message, box, button;
  int i;

  shell = XtVaCreatePopupShell (name, transientShellWidgetClass,
				parent, XtNinput, True, NULL);
  form = XtCreateManagedWidget ("form", formWidgetClass,
				shell, NULL, 0);

  message = XtVaCreateManagedWidget ("message", labelWidgetClass, form,
				     XtNleft, XtChainLeft,
				     XtNright, XtChainRight,
				     XtNtop, XtChainTop,
				     XtNbottom, XtChainBottom,
				     NULL);

  box = XtVaCreateManagedWidget ("buttonBox", boxWidgetClass, form,
				 XtNorientation, XtorientHorizontal,
				 XtNfromVert, message, NULL);

  for (i = 0; whichButtons && i < XtNumber(buttons); i++, whichButtons >>= 1)
    if (whichButtons & 1) {
      choices[i].chosen = &chosen;
      choices[i].choose = buttons[i].choice;

      button = XtCreateManagedWidget (buttons[i].name, commandWidgetClass, box,
				      NULL, 0);
      XtAddCallback (button, XtNcallback, (XtCallbackProc) buttonCallback,
		     (XtPointer) &choices[i]);
    }

  Popup (shell, parent);

  {
    XtAppContext appContext = XtWidgetToApplicationContext (shell);
    while (chosen == btUnspecified && XtIsRealized (shell))
      XtAppProcessEvent (appContext, XtIMAll);
  }

  XtDestroyWidget (shell);

  return chosen;
}

/* Ask an integer value in the specified dialog */
Boolean Ask (Widget parent, const char *name, long *value, long min, long max)
{
  ChosenButton choices[2];
  Choice chosen = btUnspecified;
  Widget shell, form, question, answer, info, box, button;
  char answertext[20];
  sprintf (answertext, "%ld", *value);

  shell = XtVaCreatePopupShell (name, transientShellWidgetClass,
				parent, XtNinput, True, NULL);
  form = XtCreateManagedWidget ("form", formWidgetClass,
				shell, NULL, 0);

  question = XtVaCreateManagedWidget ("question", labelWidgetClass, form,
				     XtNleft, XtChainLeft,
				     XtNright, XtChainRight,
				     XtNtop, XtChainTop,
				     XtNbottom, XtChainBottom,
				     NULL);

  answer = XtVaCreateManagedWidget ("answer", asciiTextWidgetClass, form,
				    XtNfromHoriz, question,
				    XtNuseStringInPlace, True,
				    XtNstring, answertext,
				    XtNlength, sizeof answertext - 1,
				    XtNeditType, XawtextEdit,
				    NULL);

  if (min > -INT_MAX || max < INT_MAX) {
    char hint[40];
    if (min > -INT_MAX && max < INT_MAX)
      sprintf (hint, "(%ld - %ld)", min, max);
    else if (min > -INT_MAX)
      sprintf (hint, "(%ld -)", min);
    else
      sprintf (hint, "(- %ld)", max);

    info = XtVaCreateManagedWidget ("info", labelWidgetClass, form,
				    XtNfromHoriz, answer,
				    XtNlabel, hint, NULL);
  }

  box = XtVaCreateManagedWidget ("buttonBox", boxWidgetClass, form,
				 XtNorientation, XtorientHorizontal,
				 XtNfromVert, question, NULL);

  choices[0].chosen = choices[1].chosen = &chosen;
  choices[0].choose = buttons[btOk].choice;
  choices[1].choose = buttons[btCancel].choice;
  button = XtCreateManagedWidget (buttons[btOk].name, commandWidgetClass,
				  box, NULL, 0);
  XtAddCallback (button, XtNcallback, (XtCallbackProc) buttonCallback,
		 (XtPointer) &choices[0]);
  button = XtCreateManagedWidget (buttons[btCancel].name, commandWidgetClass,
				  box, NULL, 0);
  XtAddCallback (button, XtNcallback, (XtCallbackProc) buttonCallback,
		 (XtPointer) &choices[1]);

  Popup (shell, parent);

  while (XtIsRealized (shell)) {
    XtAppContext appContext = XtWidgetToApplicationContext (shell);

    while (chosen == btUnspecified && XtIsRealized (shell))
      XtAppProcessEvent (appContext, XtIMAll);

    if (chosen == btCancel || !XtIsRealized (shell))
      break;

    if (chosen == btOk) {
      /* parse the value input by the user */

      char *endptr;
      long try = strtol (answertext, &endptr, 10);

      if (*answertext && !*endptr &&
	  try >= min && try <= max) {
	*value = try;
	break;
      }
    }

    XBell (XtDisplay (shell), 0);
    chosen = btUnspecified;
  }

  XtDestroyWidget (shell);

  return chosen == btOk;
}

/* The button callback function */
static void buttonCallback(Widget w, ChosenButton *c, XtPointer callData)
{
  *c->chosen = c->choose;
}
