/*
** Convert X11 events to a keyboard and joystick matrix
**
** Copyright  1997-1998 Marko Mkel, see COPYING.LIB for details
*/

#include <X11/Intrinsic.h>

#define XK_LATIN1 /* for the sterling pound sign */
#include <X11/keysym.h>

#include <stdlib.h>
#include <string.h>

#include "keyboard.h"

/* Application resources */
#include "init.h"

/* These codes are hopefully unused */
typedef enum { XK_Map_Shift_L = 1, XK_Map_Shift_R } MappedKeysyms;

#define PLAIN	0        /* normal key, shift allowed */
#define NOSHIFT	(1 << 0) /* no shift key allowed */
#define USE_S_L	(1 << 1) /* use left shift */
#define USE_S_R	(1 << 2) /* use right shift */
#define IS_S_L	(1 << 3) /* the key is left shift */
#define IS_S_R	(1 << 4) /* the key is right shift */

#define REGULAR(row,col,sym) {XK_ ## sym,row,1<<col,PLAIN}
#define SPECIAL(row,col,sym) {XK_ ## sym,row,1<<col,NOSHIFT}
#define L_SHIFT(row,col,sym) {XK_ ## sym,row,1<<col,USE_S_L}
#define R_SHIFT(row,col,sym) {XK_ ## sym,row,1<<col,USE_S_R}
#define OTHER(row,col,sym,shift) {XK_ ## sym,row,1<<col,shift}

#ifndef XtNumber
#  define XtNumber(array) (sizeof(array)/sizeof(array[0]))
#endif

static struct {
  KeySym sym;
  unsigned row;
  unsigned col;
  unsigned shift;
} Keymap [] = {
  SPECIAL(0, 0, Delete),
  SPECIAL(0, 0, BackSpace),
  L_SHIFT(0, 0, Insert),
  REGULAR(0, 1, Return),
  REGULAR(0, 1, KP_Enter),
  REGULAR(0, 2, Right),
  L_SHIFT(0, 2, Left),
  REGULAR(0, 3, F7),
  L_SHIFT(0, 3, F8),
  REGULAR(0, 4, F5),
  L_SHIFT(0, 4, F6),
  REGULAR(0, 5, F3),
  L_SHIFT(0, 5, F4),
  REGULAR(0, 6, F1),
  L_SHIFT(0, 6, F2),
  REGULAR(0, 7, Down),
  L_SHIFT(0, 7, Up),

  SPECIAL(1, 0, 3),
  SPECIAL(1, 0, KP_3),
  R_SHIFT(1, 0, numbersign),
  REGULAR(1, 1, w),
  R_SHIFT(1, 1, W),
  REGULAR(1, 2, a),
  R_SHIFT(1, 2, A),
  SPECIAL(1, 3, 4),
  SPECIAL(1, 3, KP_4),
  R_SHIFT(1, 3, dollar),
  REGULAR(1, 4, z),
  R_SHIFT(1, 4, Z),
  REGULAR(1, 5, s),
  R_SHIFT(1, 5, S),
  REGULAR(1, 6, e),
  R_SHIFT(1, 6, E),
  REGULAR(1, 7, Map_Shift_L),
  OTHER  (1, 7, Shift_L, IS_S_L),

  SPECIAL(2, 0, 5),
  SPECIAL(2, 0, KP_5),
  R_SHIFT(2, 0, percent),
  REGULAR(2, 1, r),
  R_SHIFT(2, 1, R),
  REGULAR(2, 2, d),
  R_SHIFT(2, 2, D),
  SPECIAL(2, 3, 6),
  SPECIAL(2, 3, KP_6),
  L_SHIFT(2, 3, ampersand),
  REGULAR(2, 4, c),
  R_SHIFT(2, 4, C),
  REGULAR(2, 5, f),
  R_SHIFT(2, 5, F),
  REGULAR(2, 6, t),
  R_SHIFT(2, 6, T),
  REGULAR(2, 7, x),
  R_SHIFT(2, 7, X),

  SPECIAL(3, 0, 7),
  SPECIAL(3, 0, KP_7),
  L_SHIFT(3, 0, apostrophe),
  L_SHIFT(3, 0, acute),
  L_SHIFT(3, 0, dead_acute),
  REGULAR(3, 1, y),
  L_SHIFT(3, 1, Y),
  REGULAR(3, 2, g),
  R_SHIFT(3, 2, G),
  SPECIAL(3, 3, 8),
  SPECIAL(3, 3, KP_8),
  L_SHIFT(3, 3, parenleft),
  REGULAR(3, 4, b),
  R_SHIFT(3, 4, B),
  REGULAR(3, 5, h),
  L_SHIFT(3, 5, H),
  REGULAR(3, 6, u),
  L_SHIFT(3, 6, U),
  REGULAR(3, 7, v),
  R_SHIFT(3, 7, V),

  SPECIAL(4, 0, 9),
  SPECIAL(4, 0, KP_9),
  L_SHIFT(4, 0, parenright),
  REGULAR(4, 1, i),
  L_SHIFT(4, 1, I),
  REGULAR(4, 2, j),
  L_SHIFT(4, 2, J),
  SPECIAL(4, 3, 0),		/* There is no mapping for the shifted key! */
  SPECIAL(4, 3, KP_0),		/* There is no mapping for the shifted key! */
  REGULAR(4, 4, m),
  L_SHIFT(4, 4, M),
  REGULAR(4, 5, k),
  L_SHIFT(4, 5, K),
  REGULAR(4, 6, o),
  L_SHIFT(4, 6, O),
  REGULAR(4, 7, n),
  L_SHIFT(4, 7, N),

  SPECIAL(5, 0, plus),
  SPECIAL(5, 0, KP_Add),
  L_SHIFT(5, 0, braceleft),
  REGULAR(5, 1, p),
  L_SHIFT(5, 1, P),
  REGULAR(5, 2, l),
  L_SHIFT(5, 2, L),
  SPECIAL(5, 3, minus),
  SPECIAL(5, 3, KP_Subtract),
  L_SHIFT(5, 3, braceright),
  L_SHIFT(5, 3, bar),
  SPECIAL(5, 4, period),
  SPECIAL(5, 4, KP_Separator),
  L_SHIFT(5, 4, greater),
  SPECIAL(5, 5, colon),
  L_SHIFT(5, 5, bracketleft),
  SPECIAL(5, 6, at),		/* The shifted key does not produce ASCII! */
  SPECIAL(5, 7, comma),
  L_SHIFT(5, 7, less),

#ifdef XK_sterling
  SPECIAL(6, 0, sterling),	/* The shifted key does not produce ASCII! */
#endif
  SPECIAL(6, 0, backslash),	/* Sterling Pound () for ASCII keyboards */
  SPECIAL(6, 1, asterisk),
  SPECIAL(6, 1, KP_Multiply),
  L_SHIFT(6, 1, grave),
  L_SHIFT(6, 1, dead_grave),
  SPECIAL(6, 2, semicolon),
  L_SHIFT(6, 2, bracketright),
  REGULAR(6, 3, Home),
  REGULAR(6, 4, Map_Shift_R),
  OTHER  (6, 4, Shift_R, IS_S_R),
  REGULAR(6, 5, equal),
  SPECIAL(6, 6, asciicircum),
  SPECIAL(6, 6, dead_circumflex),
  L_SHIFT(6, 6, asciitilde),
  L_SHIFT(6, 6, dead_tilde),
  SPECIAL(6, 7, slash),
  SPECIAL(6, 7, KP_Divide),
  L_SHIFT(6, 7, question),

  SPECIAL(7, 0, 1),
  SPECIAL(7, 0, KP_1),
  R_SHIFT(7, 0, exclam),
  SPECIAL(7, 1, underscore),	/* There is no mapping for the shifted key! */
  REGULAR(7, 2, Control_L),
  REGULAR(7, 2, Control_R),
  SPECIAL(7, 3, 2),
  SPECIAL(7, 3, KP_2),
  R_SHIFT(7, 3, quotedbl),
  REGULAR(7, 4, space),
  REGULAR(7, 5, Alt_L),		/* Commodore key */
  REGULAR(7, 5, Alt_R),
  REGULAR(7, 5, Meta_L),
  REGULAR(7, 5, Meta_R),
  REGULAR(7, 6, q),
  R_SHIFT(7, 6, Q),
  REGULAR(7, 7, Escape),	/* STOP key */

  REGULAR(7, 7, VoidSymbol)	/* end of list */
};
#undef REGULAR
#undef SPECIAL
#undef L_SHIFT
#undef R_SHIFT
#undef OTHER

typedef enum {
  NEUTRAL = 0,
  North, South, NORTHSOUTH,
  West, Northwest, Southwest, NORTHSOUTHWEST,
  East, Northeast, Southeast, NORTHSOUTHEAST,
  WESTEAST, NORTHWESTEAST, SOUTHWESTEAST, NORTHSOUTHWESTEAST,
  Fire,
  FireNorth, FireSouth, FireNORTHSOUTH,
  FireWest, FireNorthwest, FireSouthwest, FireNORTHSOUTHWEST,
  FireEast, FireNortheast, FireSoutheast, FireNORTHSOUTHEAST,
  FireWESTEAST, FireNORTHWESTEAST, FireSOUTHWESTEAST, FireNORTHSOUTHWESTEAST
} JoystickDirection;

static KeySym stick1[32], stick2[32];

/* Change a mapping */
static void MapKey (KeySym old, KeySym new);

/* Look up a key */
static int LookupKey (KeySym key);

/* Look up a joystick key */
static int LookupStick (KeySym *stick, KeySym key);

/* Indexes to the shift keys */
static int shiftL, shiftR;

/* Initialize the keyboard */
void InitKeyboard ()
{
  KeySym sym;

  /* Bind the keyboard joysticks */
#define BINDSTICK(port,dir)						   \
  if (NoSymbol != (sym=XStringToKeysym (appResources.joy ## port ## dir))) \
    stick ## port[dir] = sym
#define BINDSTICKS(dir) BINDSTICK(1,dir); BINDSTICK(2,dir)

  BINDSTICKS(North);
  BINDSTICKS(South);
  BINDSTICKS(West);
  BINDSTICKS(Northwest);
  BINDSTICKS(Southwest);
  BINDSTICKS(East);
  BINDSTICKS(Northeast);
  BINDSTICKS(Southeast);
  BINDSTICKS(Fire);

#undef BINDSTICK
#undef BINDSTICKS

  /* Set the shift key indices */
  shiftL = LookupKey (XK_Shift_L);
  shiftR = LookupKey (XK_Shift_R);

  /* Bind the additional shift keys */
  if (NoSymbol != (sym = XStringToKeysym (appResources.keyShiftL))) {
    MapKey (sym, NoSymbol);
    MapKey (XK_Map_Shift_L, sym);
  }
  if (NoSymbol != (sym = XStringToKeysym (appResources.keyShiftR))) {
    MapKey (sym, NoSymbol);
    MapKey (XK_Map_Shift_R, sym);
  }
}

/* Clean up the data structures */
void CloseKeyboard()
{
  /* dummy arguments */
  unsigned joy1, joy2;
  unsigned char keys[8];
  /* free the ActiveKeys structure */
  ReadKeyboard (NoSymbol, 0, LeaveNotify, &joy1, &joy2, keys);
}

/* Process events related with the keyboard */
void ReadKeyboard (KeySym key, unsigned int keycode,
		   int type, unsigned *joy1, unsigned *joy2,
		   unsigned char *keys)
{
  static Boolean ShiftL = False, ShiftR = False;
  static Boolean RawShiftL = False, RawShiftR = False;
  Boolean pressed = type == KeyPress;
  int i;

  /* Structure for memorizing the keypresses */
  static struct ActiveKeys {
    struct ActiveKeys *next;
    KeySym key;
    unsigned int keycode;
  } *activeKeys = NULL;

  /* Clean up at LeaveNotify */
  if (type == LeaveNotify) {
    /* Clear active keys */
    while (activeKeys) {
      struct ActiveKeys *ak = activeKeys->next;
      free (activeKeys);
      activeKeys = ak;
    }

    *joy1 = *joy2 = 0;
    RawShiftL = RawShiftR = ShiftL = ShiftR = False;
    memset (keys, 0, 8);
    return;
  }

  /* Handle active keys.  According to a comment in the VICE source
   * code, X11 can sometimes give the KeyPress event with the shifted
   * KeySym, and the KeyRelease one with the same KeySym unshifted.
   * The keycode field is equal in both events. */

  /* Not turning off the autorepeat fixes some of the behaviour, but
   * e.g.  if XK_mode_shift+XK_3 is defined to be XK_sterling, then
   * the following sequence would normally cause problems (the
   * Sterling pound character would autorepeat forever):
   * <KeyPress XK_mode_shift><KeyPress XK_sterling><KeyRelease XK_mode_shift>.
   * Storing the keycodes fixes these problems. */

  if (pressed) { /* Add the key if it is not on the list. */
    struct ActiveKeys **ak = &activeKeys;
    while (*ak && (*ak)->keycode != keycode)
      ak = &(*ak)->next;

    if (*ak) {
      (*ak)->key = key;
    } else {
      *ak = malloc (sizeof (**ak));
      (*ak)->next = NULL;
      (*ak)->key = key;
      (*ak)->keycode = keycode;
    }
  }
  else { /* Use the same keysym that was in the KeyPress message */
    struct ActiveKeys **ak = &activeKeys;
    while (*ak && (*ak)->keycode != keycode)
      ak = &(*ak)->next;

    if (*ak) {
      struct ActiveKeys *k = *ak;
      key = (*ak)->key;
      *ak = (*ak)->next;
      free (k);
    }
  }

  /* Handle joystick emulation keys */

  if (appResources.joy1 && (0 <= (i = LookupStick (stick1, key)))) {
    if (pressed)
      *joy1 |= i;
    else
      *joy1 &= ~i;
    return;
  }

  if (appResources.joy2 && (0 <= (i = LookupStick (stick2, key)))) {
    if (pressed)
      *joy2 |= i;
    else
      *joy2 &= ~i;
    return;
  }

  /* Handle normal keys */

  if ((i = LookupKey (key)) < 0)
    return;

#define SETKEY(k) keys[Keymap[k].row] |= Keymap[k].col
#define RESETKEY(k) keys[Keymap[k].row] &= ~ Keymap[k].col

  /* Handle the shift keys */

  if (Keymap[i].shift == IS_S_L) {
    ShiftL = pressed;
    if (pressed) SETKEY(i); else RESETKEY(i);
  }
  else if (Keymap[i].shift == IS_S_R) {
    ShiftR = pressed;
    if (pressed) SETKEY(i); else RESETKEY(i);
  }
  else if (Keymap[i].row == Keymap[shiftL].row &&
	   Keymap[i].col == Keymap[shiftL].col) {
    RawShiftL = pressed;
    if (pressed) SETKEY(i); else RESETKEY(i);
  }
  else if (Keymap[i].row == Keymap[shiftR].row &&
	   Keymap[i].col == Keymap[shiftR].col) {
    RawShiftR = pressed;
    if (pressed) SETKEY(i); else RESETKEY(i);
  }
  else if (pressed) {
    /* Normal key pressed */
    SETKEY(i);

    /* Set the shift keys */
    if (!ShiftL || Keymap[i].shift == NOSHIFT || Keymap[i].shift == USE_S_R)
      RESETKEY(shiftL);
    else if (ShiftL)
      SETKEY(shiftL);
    if (RawShiftL)
      SETKEY(shiftL);

    if (!ShiftR || Keymap[i].shift == NOSHIFT || Keymap[i].shift == USE_S_L)
      RESETKEY(shiftR);
    else if (ShiftR)
      SETKEY(shiftR);
    if (RawShiftR)
      SETKEY(shiftR);

    if (Keymap[i].shift == USE_S_L)
      SETKEY(shiftL);

    if (Keymap[i].shift == USE_S_R)
      SETKEY(shiftR);
  }
  else {
    /* Normal key released */
    RESETKEY(i);

    /* Reset the shift keys */
    if (ShiftL)
      SETKEY(shiftL);
    else
      RESETKEY(shiftR);

    if (ShiftR)
      SETKEY(shiftL);
    else
      RESETKEY(shiftR);
  }
}

/* Change a mapping */
static void MapKey (KeySym old, KeySym new)
{
  int i;
  for (i = 0; i < XtNumber(Keymap); i++)
    if (Keymap[i].sym == old) {
      Keymap[i].sym = new;
      return;
    }
}

/* Look up a key */
static int LookupKey (KeySym key)
{
  int i;

  for (i = 0; Keymap[i].sym != XK_VoidSymbol; i++)
    if (Keymap[i].sym == key)
      return i;

  return -1;
}

/* Look up a joystick key */
static int LookupStick (KeySym *stick, KeySym key)
{
  int i;

  for (i = 0; i < 32; i++)
    if (stick[i] == key)
      return i;

  return -1;
}
