/*
** Menu implementation
**
** Copyright  1997 Marko Mkel, see COPYING.LIB for details
*/

#include <X11/Intrinsic.h>
#include <X11/StringDefs.h>
#include <X11/Shell.h>

#include <X11/Xaw/SimpleMenu.h>
#include <X11/Xaw/SmeBSB.h>
#include <X11/Xaw/SmeLine.h>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "menu.h"
#include "util.h"

#define ActionProc(name)\
static void name(Widget w, XEvent *event, String *params, Cardinal *num_params)

/* Action procedure for popping up and down SimpleMenu submenus */
ActionProc(positionSubmenuAction);
/* Pop up a menu */
ActionProc(popupMenuAction);

#define CallbackProc(name)\
static void name(Widget w, XtPointer clientData, XtPointer callData)

/* SimpleMenu popdownCallback handler */
CallbackProc(popdownSubmenuCallback);

/* SimpleMenu callback handler */
static void doMenu (Widget w, MenuEntry *me, XtPointer callData);

/* Create the pixmaps used with some menu entries */
static void createPixmaps (Widget parent);

/* Create a menu */
static Widget createMenu (Widget parent, const char *name,
			  MenuEntry *entries, int nentries);

/* Get the Menu item associated with a SimpleMenu widget */
static Menu *getMenu (Widget menu);
/* Get the MenuEntry item associated with the SimpleMenu and SmeBSB widgets */
static MenuEntry *getMenuEntry (Widget menu, Widget entry);

static Pixmap pixChecked = XtUnspecifiedPixmap,
  pixUnchecked = XtUnspecifiedPixmap,
  pixSub = XtUnspecifiedPixmap;

/* All menus managed by this module */

static Menu **allMenus = NULL;
Cardinal nAllMenus = 0;

/* Create the menus */
Boolean CreateMenus (Widget parent, Menu **menus)
{
  Cardinal i;
  static XtActionsRec actions[] = {
    { "PositionSubmenu", positionSubmenuAction },
    { "PopupMenu", popupMenuAction },
  };

  XtAppAddActions (XtWidgetToApplicationContext (parent),
		   actions, XtNumber (actions));
  XtRegisterGrabAction (popupMenuAction, True,
			ButtonPressMask | ButtonReleaseMask,
			GrabModeAsync, GrabModeAsync);
  createPixmaps (parent);

  for (i = 0; menus[i]; i++)
    if (!(menus[i]->widget =
	  createMenu (parent, menus[i]->name,
		      menus[i]->entries, menus[i]->num_entries)))
      return False;

  if (!i)
    return True;

  /* Keep track on the menus */

  if (!allMenus)
    allMenus = malloc (i * sizeof(*allMenus));
  else
    allMenus = realloc (allMenus, (nAllMenus + i) * sizeof(*allMenus));

  if (!allMenus) return False;

  memcpy (&allMenus[nAllMenus], menus, i * sizeof(*allMenus));
  nAllMenus += i;

  return True;
}

/* Update the checkable menu entries */
void UpdateMenus ()
{
  Arg arg;
  Cardinal i, j;

  for (i = 0; i < nAllMenus; i++) {
    MenuEntry *me = allMenus[i]->entries;

    for (j = allMenus[i]->num_entries; j--; me++) {
      if (me->type == menuCheckable && me->function) {
	XtSetArg (arg, XtNleftBitmap,
		  (*me->function) (False) ? pixChecked : pixUnchecked);
	XtSetValues (me->widget, &arg, 1);
      }
    }
  }
}

/*
** Static functions
*/

/* Create a menu */
static Widget createMenu (Widget parent, const char *name,
			  MenuEntry *entries, int nentries)
{
  Widget m;

  static XtCallbackRec cb[2] = {
    { (XtCallbackProc) doMenu, NULL },
    { NULL, NULL }
  };

  static Arg args[] = {
    { XtNcallback, (XtArgVal) cb },
    { XtNleftBitmap, (XtArgVal) NULL },
    { XtNrightBitmap, (XtArgVal) NULL },
  };

  args[2].value = (XtArgVal) pixSub;

  if (!(m = XtCreatePopupShell (name, simpleMenuWidgetClass, parent, NULL, 0)))
    return NULL;

  XtOverrideTranslations (m, XtParseTranslationTable
			  ("<BtnMotion>: highlight() PositionSubmenu()\n"));
  XtAddCallback(m, XtNpopdownCallback, popdownSubmenuCallback, NULL);

  for (; nentries > 0; nentries--, entries++) {
    cb[0].closure = entries;

    if (entries->type == menuCheckable)
      args[1].value = (XtArgVal) ((entries->function &&
				   (*entries->function) (False)) ?
				  pixChecked : pixUnchecked);
    else
      args[1].value = (XtArgVal) NULL;

    entries->widget = XtCreateManagedWidget
      (entries->name,
       entries->type == menuSeparator ? smeLineObjectClass : smeBSBObjectClass,
       m, args, entries->submenu ? 3 : 2);
  }

  return m;
}

/* Create the pixmaps used with some menu entries */
static void createPixmaps (Widget parent)
{
#define checked_width 11
#define checked_height 11
  static unsigned char checked_bits[] = {
    0xff,0x07,0x03,0x06,0x05,0x05,0x89,0x04,0x51,0x04,0x21,0x04,0x51,0x04,0x89,
    0x04,0x05,0x05,0x03,0x06,0xff,0x07
  };

#define unchecked_width 11
#define unchecked_height 11
  static unsigned char unchecked_bits[] = {
    0xff,0x07,0x01,0x04,0x01,0x04,0x01,0x04,0x01,0x04,0x01,0x04,0x01,0x04,0x01,
    0x04,0x01,0x04,0x01,0x04,0xff,0x07
  };

#define sub_width 11
#define sub_height 11
  static unsigned char sub_bits[] = {
    0x01,0x00,0x07,0x00,0x1f,0x00,0x7f,0x00,0xff,0x01,0xff,0x07,0xff,0x01,0x7f,
    0x00,0x1f,0x00,0x07,0x00,0x01,0x00
  };

  if (pixChecked == XtUnspecifiedPixmap)
    pixChecked = XCreateBitmapFromData (XtDisplay(parent),
					RootWindowOfScreen(XtScreen(parent)),
					(char *)checked_bits,
					checked_width, checked_height);

  if (pixUnchecked == XtUnspecifiedPixmap)
    pixUnchecked = XCreateBitmapFromData (XtDisplay(parent),
					  RootWindowOfScreen(XtScreen(parent)),
					  (char *)unchecked_bits,
					  unchecked_width, unchecked_height);

  if (pixSub == XtUnspecifiedPixmap)
    pixSub = XCreateBitmapFromData (XtDisplay(parent),
				    RootWindowOfScreen(XtScreen(parent)),
				    (char *)sub_bits,
				    sub_width, sub_height);
}

/* SimpleMenu callback handler */
static void doMenu (Widget w, MenuEntry *me, XtPointer callData)
{
  Display *display = XtDisplayOfObject (w);

  /* Pop down all menus */
  {
    Cardinal i;

    for (i = 0; i < nAllMenus; i++)
      XtPopdown (allMenus[i]->widget);
  }

  if (me->type == menuCheckable) {
    if (me->function && (*me->function) (True))
      UpdateMenus ();
    else
      XBell (display, 0);
  }
  else if (me->type == menuNormal) {
    if (me->function) {
      if (!(*me->function) (True))
	XBell (display, 0);
    }
    else if (!me->submenu)
      XBell (display, 0);
  }

  if (me->submenu) {
    Menu *m = getMenu (XtParent (w));
    if (m && m->activeSubmenu) {
      fprintf (stderr,
	       "Congratulations, you found a bug in the menu code!\n"
	       "The menus just stopped working.  However, you can still\n"
	       "use other parts of the program.\n");
    }
  }
}

/* Get the Menu item associated with a SimpleMenu widget */
static Menu *getMenu (Widget menu)
{
  Cardinal i;

  for (i = 0; i < nAllMenus; i++)
    if (allMenus[i]->widget == menu)
      return allMenus[i];

  return NULL;
}

/* Get the MenuEntry item associated with the SimpleMenu and SmeBSB widgets */
static MenuEntry *getMenuEntry (Widget menu, Widget entry)
{
  Menu *m = getMenu (menu);
  MenuEntry *entries = m ? m->entries : NULL;
  Cardinal i;

  if (!entries || !entry)
    return NULL;

  for (i = 0; i < m->num_entries; i++)
    if (entries[i].widget == entry)
      return &entries[i];

  return NULL;
}

/* Action procedure for popping up and down SimpleMenu submenus */
ActionProc(positionSubmenuAction)
{
  static Widget current_menu = NULL;
  Widget new_menu, entry = XawSimpleMenuGetActiveEntry (w);

  {
    MenuEntry *me = getMenuEntry (w, entry);
    new_menu = (me && me->submenu) ? me->submenu->widget : NULL;
  }

  {
    Menu *m = getMenu(w);

    if (m->activeSubmenu && m->activeSubmenu != new_menu)
      XtPopdown (m->activeSubmenu);

    m->activeSubmenu = new_menu;
  }

  if (new_menu && new_menu != current_menu) {
    /* Position the submenu */
    Position ex, ey, x, y;
    Dimension ewidth, width, height;

    /* Get the active menu item geometry */
    XtVaGetValues (entry, XtNx, &ex, XtNy, &ey, XtNwidth, &ewidth, NULL);
    XtTranslateCoords(w, ex, ey, &ex, &ey);

    /* Get the new menu dimensions */
    XtVaGetValues (new_menu, XtNwidth, &width, XtNheight, &height, NULL);

    x = ex + ewidth;
    y = ey;

    /* Make the menu fit on the screen */
    ValidateCoordinates (XtDisplayOfObject (w), &x, &y, width, height,
			 ewidth + 2, 0);

    XtVaSetValues (new_menu, XtNx, x, XtNy, y, NULL);
    XtPopup (new_menu, XtGrabNonexclusive);
  }

  current_menu = new_menu;
}

/* Pop up a menu */
ActionProc(popupMenuAction)
{
  if (*num_params == 1) {
    XtCallActionProc (w, "XawPositionSimpleMenu", event, params, 1);
    XtCallActionProc (w, "MenuPopup", event, params, 1);
  }
}

/* SimpleMenu popdownCallback handler */
CallbackProc(popdownSubmenuCallback)
{
  Cardinal i;
  Menu *m = getMenu (w);

  /* Destroy the submenu of this menu, if any is active */
  if (m && m->activeSubmenu) {
    XtAddGrab (m->activeSubmenu, False, False); /* suppress a warning */
    XtPopdown (m->activeSubmenu);
  }

  /* Tell any parent menu that this menu has been popped down */
  for (i = 0; i < nAllMenus; i++)
    if (allMenus[i]->activeSubmenu == w)
      allMenus[i]->activeSubmenu = NULL;
}
