/*
** User interface for the X11 version of CCS64
**
** Copyright  1997-1998 Marko Mkel, see COPYING.LIB for details
*/

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>

#include <limits.h>

#ifndef NO_ZLIB
#  include <zlib.h>
#else
  typedef FILE *gzFile;
#  define gzopen(name,mode) fopen(name,mode)
#  define gzclose(file) fclose(file)
#  define gzread(file,buf,len) fread(buf,1,len,file)
#endif

#include <X11/Intrinsic.h>
#include <X11/StringDefs.h>
#include <X11/Shell.h>

#include <X11/keysym.h>

extern "C" {
#include "menu.h"
#include "init.h"
#include "keyboard.h"
#include "dialog.h"
#include "filedialog.h"
}

/* The basic geometry (PAL C64 system) */
#define CYCLES_PER_LINE 63
#define LINES 312
#define COLUMNS CYCLES_PER_LINE * 8

#define CLOCKSPEED 985248

#define FIXLEFTCOLUMN	"128"
#define FIXTOPLINE	"51"

#define CYCLES_PER_FRAME (LINES * CYCLES_PER_LINE)

/* The application class */
#define CLASS "CCS64"

/* Fallback resources and menu structure */
#include "ui.h"

/*
** Main program
*/

#include <sys/time.h>
#include <sys/types.h>
#ifdef _AIX
#  include <sys/select.h>
#endif
#include "emulator.h"
#include "c1541.h"

int main (int argc, char **argv);

/* X callback functions */
EventHandler(HandleKeys);
EventHandler(HandleMotion);

/* Turn the drive LED on/off */
static void setDriveLed (Boolean state);

/* Main emulation loop */
static void mainloop (XtAppContext *appContext);

/* Read a ROM file */
static Boolean readRom (String name, unsigned char *buf, int len);

/* Read a program file or a disk image to the emulator's memory */
static Boolean readFile (String name, Boolean UsePath);

/* Emulate the videochip for the specified amount of raster lines
   and draw the graphics */
static void drawGraphics (unsigned rasters);

/* Raster counter (for drawGraphics) */
static unsigned RasterCounter = 0;

/* Special keys */
static KeySym restoreKey;

static C64 c64;

int main (int argc, char **argv)
{
  static unsigned char kernal[8192], basic[8192], character[4096];
  static unsigned char exrom[8192], drive[16384];

  XtAppContext appContext;

  if (!InitializeXt (&appContext, CLASS, fallbackRes,
		     menus, COLUMNS, LINES, &argc, argv))
    return 1;

  XtAddEventHandler (TopLevel, KeyPressMask | KeyReleaseMask | LeaveWindowMask,
		     False, HandleKeys, NULL);
  XtAddEventHandler (TopLevel, PointerMotionMask, False, HandleMotion, NULL);

  if (!readRom (appResources.kernalRom, kernal, sizeof kernal) ||
      !readRom (appResources.basicRom, basic, sizeof basic) ||
      !readRom (appResources.characterRom, character, sizeof character) ||
      !readRom (appResources.driveRom, drive, sizeof drive)) {
    XtDestroyApplicationContext (appContext);
    return 2;
  }

  if (*appResources.exRom &&
      !readRom (appResources.exRom, exrom, sizeof exrom)) {
    XtDestroyApplicationContext (appContext);
    return 2;
  }

  c64.EnableC1541 (appResources.emulate1541);
  c64.EnableCartridge (appResources.exrom);
  c64.SetRom (0xe000, sizeof kernal, kernal);
  c64.SetRom (0xa000, sizeof basic, basic);
  c64.SetRom (0xd000, sizeof character, character);
  c64.SetRom (0x8000, sizeof exrom, exrom);
  c64.c1541->SetRom (0xc000, sizeof drive, drive);
  c64.c1541->Reset ();
  c64.Reset ();

  if (argc > 1 || *appResources.commandString) {
    /* Let the system finish the reset routine */
    int i;
    for (i = 0; i < 2142100; i += CYCLES_PER_FRAME)
      c64.Run6510 (0);
  }

  /* Load any files specified by the user */
  {
    int i;
    for (i = 1; i < argc; i++)
      readFile (argv[i], True);
  }

  InitKeyboard ();
  restoreKey = XStringToKeysym (appResources.keyRestore);
  UpdateMenus ();
  setDriveLed (False); /* Turn off the drive LED indicator */
  mainloop (&appContext);
  setDriveLed (False); /* Turn off the drive LED indicator */
  Cleanup ();
  return 0;
}

/* Keyboard handler */
EventHandler(HandleKeys)
{
  static unsigned joy1 = 0, joy2 = 0;
  static unsigned char keys[8];

  if (w != TopLevel)
    return;

  switch (event->type) {
    KeySym key;
  case KeyPress:
  case KeyRelease:
    XLookupString (&event->xkey, NULL, 0, &key, NULL);

    if (event->type == KeyPress && key == restoreKey) {
      c64.SetRestoreKey ();
      return;
    }

    ReadKeyboard (key, event->xkey.keycode, event->type, &joy1, &joy2, keys);
    break;

  case LeaveNotify:
    ReadKeyboard (NoSymbol, 0, event->type, &joy1, &joy2, keys);
    break;
  }

  c64.SetJoystick (0, joy1);
  c64.SetJoystick (1, joy2);
  c64.SetKeys (keys);
}

EventHandler(HandleMotion)
{
  Dimension x, y;
  Arg args[2];
  
  if (w != TopLevel)
    return;

  XtSetArg (args[0], XtNwidth, &x);
  XtSetArg (args[1], XtNheight, &y);
  XtGetValues (w, args, XtNumber (args));

  if (!x || !y)
    return;

  x = 255 - (255 * event->xmotion.x / x);
  y = 255 - (255 * event->xmotion.y / y);

  c64.SetPaddles (0, x, y);
}

/* Turn the drive LED on/off */
static void setDriveLed (Boolean state) {
  if (appResources.driveLed) {
    XKeyboardControl values;
    values.led_mode = state;
    values.led = appResources.driveLed;
    XChangeKeyboardControl (XtDisplay (TopLevel),
			    KBLed | KBLedMode, &values);
  }
}

/* Command string (automatically inserted to the keyboard buffer) */
static char *command;

/* Main emulation loop */
static void mainloop (XtAppContext *appContext)
{
  char *title;
  unsigned offset;
  static int oldpercent = -1;

  /* Set up the things for the select() call (delays) */
  fd_set rfds;
  int xconn = XConnectionNumber (XtDisplay (TopLevel));
  FD_ZERO (&rfds);

  /* Allocate room for the dynamic title string */
  {
    unsigned len;
    char *origtitle;
    /* Get the window title */
    XtVaGetValues (TopLevel, XtNtitle, &origtitle, NULL);

    offset = strlen (origtitle);
    len = strlen (appResources.pausedString) + 1;

#define DISPLAYLENGTH 20
#define DISPLAY(speed)					\
    do {						\
      sprintf (&title[offset], " (%d%%)", speed); 	\
      XtVaSetValues (TopLevel, XtNtitle, title, NULL);	\
    } while (0)

    if (len < DISPLAYLENGTH)
      len = DISPLAYLENGTH; /* extra space for the percent display */

    if (!(title = new char [len + offset]))
      return;

    strcpy (title, origtitle);
  }

  command = appResources.commandString;

  while (!closed) {
    XEvent event;

    if (appResources.paused) {
      /* Set the window title */
      strcpy (&title[offset], appResources.pausedString);
      XtVaSetValues (TopLevel, XtNtitle, title, NULL);
      oldpercent = -1;

      /* Draw the whole screen */
      drawGraphics (LINES);

      /* Process events */
      while (XtAppPending (*appContext) || appResources.paused) {
	XtAppNextEvent (*appContext, &event);
	XtDispatchEvent (&event);
	if (closed) goto Closed;
      }
    }

    while (XtAppPending (*appContext)) {
      XtAppNextEvent (*appContext, &event);
      XtDispatchEvent (&event);
    }

    /* Run the emulator for a frame */
    {
      static int frameskip = 0, frames = 0;
      static struct timeval newtimer;

      if (!timerisset (&newtimer))
	gettimeofday (&newtimer, NULL);

      newtimer.tv_usec += (appResources.speed
			   ? (((CLOCKSPEED / 10) * CYCLES_PER_FRAME / 1000) /
			      appResources.speed)
			   : 0);

      while (newtimer.tv_usec > 1000000) {
	newtimer.tv_usec -= 1000000;
	newtimer.tv_sec++;
      }

      /* Emulate the CPU for a frame */
      c64.Run6510 (0);

      /* Emulate the drive LED */
      if (appResources.driveLed) {
	static Boolean ledState = False;

	if (ledState != (c64.C1541IsEnabled () && c64.c1541->LedStatus ()))
	  setDriveLed (ledState ^= True);
      }

      /* Fill the keyboard buffer with the specified characters */
      if (*command) {
	unsigned char kbdbuflen;
	c64.GetRam (198, 1, &kbdbuflen);

	if (!kbdbuflen) {
	  unsigned char kbdbuf [10];
	  unsigned i = strlen (command);

	  kbdbuflen = (i > sizeof kbdbuf) ? sizeof kbdbuf : i;

	  /* Convert the string to PETSCII */
	  for (i = 0; i < kbdbuflen; i++)
	    if (command[i] >= 'A' && command[i] <= 'Z')
	      kbdbuf[i] = command[i] - 'A' + 0xC1;
	    else if (command[i] >= 'a' && command[i] <= 'z')
	      kbdbuf[i] = command[i] - 'a' + 0x41;
	    else if (command[i] == '\n')
	      kbdbuf[i] = 0x0d;
	    else
	      kbdbuf[i] = command[i];

	  /* Set the keyboard buffer */
	  c64.SetRam (198, 1, &kbdbuflen);
	  c64.SetRam (631, kbdbuflen, kbdbuf);

	  /* Get next text segment */
	  command += kbdbuflen;
	}
      }
      else
	command = GetCommands ();

      /* Update the screen */
      if (frameskip)
	frameskip--;
      else {
	frameskip = appResources.frameSkip;
	/* Update the screen */

	{
	  int parts = appResources.frameParts;
	  int lines = LINES / parts;
	
	  while (parts--) {
	    struct timeval timer;
	    drawGraphics (lines);
	    gettimeofday (&timer, NULL);
	    if (timercmp (&timer, &newtimer, >))
	      break;
	  }
	}

	/* Spend the rest of the time */
	for (;;) {
	  struct timeval timer;
	  gettimeofday (&timer, NULL);
	  if (!timercmp (&timer, &newtimer, <)) {
	    newtimer = timer;
	    break;
	  }

	  if (XtAppPending (*appContext)) {
	    XtAppNextEvent (*appContext, &event);
	    XtDispatchEvent (&event);
	  }
	  else {
	    struct timeval timeout;
	    timeout.tv_sec = newtimer.tv_sec - timer.tv_sec;
	    if (newtimer.tv_usec < timer.tv_usec) {
	      timeout.tv_sec--;
	      timeout.tv_usec = 1000000 + newtimer.tv_usec - timer.tv_usec;
	    }
	    else
	      timeout.tv_usec = newtimer.tv_usec - timer.tv_usec;

	    FD_SET (xconn, &rfds);
	    select (xconn + 1, &rfds, NULL, NULL, &timeout);
	  }
	}

	/* Display the statistics */
	if (frames) {
	  static struct timeval origtimer;
	  int elapsed, percent;

	  elapsed =
	    (newtimer.tv_sec - origtimer.tv_sec) * 1000000 +
	    newtimer.tv_usec - origtimer.tv_usec;

	  percent = elapsed ? 100 * CYCLES_PER_FRAME * frames / elapsed : 0;
	  origtimer = newtimer;
	  frames = 0;

	  if (oldpercent != percent)
	    DISPLAY(oldpercent = percent);
	}
      }

      frames++;
    }
  }

Closed:
  delete[] title;
}

/* Read a ROM file */
static Boolean readRom (String name, unsigned char *buf, int len)
{
  gzFile f;
  String filename;
  SubstitutionRec sr;

  sr.match = 's';
  sr.substitution = name;

  if (!(filename = XtFindFile (appResources.romPath, &sr, 1, NULL))) {
    fprintf (stderr, "Could not find file %s\n", name);
    return False;
  }

  if (!(f = gzopen (filename, "rb"))) {
  readError:
    fprintf (stderr, "Could not read file %s\n", filename);
    XtFree (filename);

    return False;
  }

  if (len != gzread (f, buf, len)) {
    gzclose (f);
    goto readError;
  }

  gzclose (f);
  XtFree (filename);
  return True;
}

/* Read a program file or a disk image to the emulator's memory */
static Boolean readFile (String name, Boolean UsePath)
{
  gzFile f;
  String filename = NULL;
  SubstitutionRec sr;
  unsigned char *buffer;
  long length;

  sr.match = 's';
  sr.substitution = name;

  if (UsePath) {
    if (!(filename = XtFindFile (appResources.filePath, &sr, 1, NULL))) {
      fprintf (stderr, "Could not find file %s\n", name);
      return False;
    }

    name = filename;
  }

  if (!(f = gzopen (name, "rb"))) {
    perror ("fopen");
  readError:
    if (f) gzclose (f);
    fprintf (stderr, "Could not read file %s\n", name);
    if (filename) XtFree (filename);

    return False;
  }

#define DISKSIZE 174848
  buffer = new unsigned char [DISKSIZE + 1];

  if (!buffer) {
    fprintf (stderr, "Out of memory.\n");
    goto readError;
  }

  length = gzread (f, buffer, DISKSIZE + 1);

  if (length == -1) {
    delete[] buffer;
    perror ("fread");
    goto readError;
  }

  if ((length < 3 || length > 0x10002) && length != DISKSIZE) {
    delete[] buffer;
    fprintf (stderr, "File length out of range: %ld\n", length);
    goto readError;
  }

  gzclose (f);
  if (filename) XtFree (filename);

  if (length == DISKSIZE)
    c64.c1541->LoadImageD64 (buffer, 35);
  else {
    unsigned start = buffer[0] | (buffer[1] << 8);
    unsigned end = start + length - 2;

    c64.SetRam (start, (end & 0xffff) - start, &buffer[2]);
    if (end > 0xffff)
      c64.SetRam (0, end & 0xffff, &buffer[2 + (end & 0xffff)]);

    /* Set the start and end addresses */
    c64.SetRam (0xac, 2, buffer);
    buffer[0] = end & 0xff;
    buffer[1] = end >> 8;
    c64.SetRam (0xae, 2, buffer);
    c64.SetRam (0x2d, 2, buffer);
    c64.SetRam (0x2f, 2, buffer);
    c64.SetRam (0x31, 2, buffer);
  }

  delete[] buffer;
  return True;
}

/* Emulate the videochip for the specified amount of rasterlines
   and draw the graphics */
static void drawGraphics (unsigned rasters)
{
  for (; rasters--; RasterCounter++, RasterCounter %= LINES) {
    (*convertLine) (c64.UpdateNextRaster (), RasterCounter);
    displayLine (RasterCounter);
  }
}

/*
** Menu handlers
*/

static Boolean do_load (Boolean flag)
{
  Boolean status = True;
  char *filename;

  if (!flag) return False;

  filename = openFile (TopLevel);

  if (filename) {
    status = readFile (filename, False);
    free (filename);
  }

  return status;
}

static Boolean do_save (Boolean flag)
{
  Boolean status = False;
  char *filename;
  unsigned start, end;

  if (!flag) return False;

  /* Read the BASIC start and end addresses */
  {
    unsigned char byte;
    c64.GetRam (0x2b, 1, &byte);
    start = byte;
    c64.GetRam (0x2c, 1, &byte);
    start |= byte << 8;
    c64.GetRam (0x2d, 1, &byte);
    end = byte;
    c64.GetRam (0x2e, 1, &byte);
    end |= byte << 8;
  }

  filename = writeFile (TopLevel, True, &start, &end);

  if (filename) {
    FILE *f;
    unsigned len = (start <= end) ? end - start : 0x10000 + end - start;
    unsigned char *buf = NULL;

    if (!(f = fopen (filename, "wb")))
      perror ("fopen");
    else if ((buf = new unsigned char [len])) {
      c64.GetRam (start, start + len > 0xffff ? 0x10000 - start : len, buf);
      if (start > end)
	c64.GetRam (0, end, &buf[0x10000 - start]);

      if (EOF == putc (start & 0xff, f) ||
	  EOF == putc (start >> 8, f))
	perror ("putc");
      else if (1 != fwrite (buf, len, 1, f))
	perror ("fwrite");
      else
	status = True;
    }

    if (f) fclose (f);
    delete[] buf;
    free (filename);
  }

  return status;
}

static Boolean do_saveImage (Boolean flag)
{
  Boolean status = False;
  char *filename;

  if (!flag) return False;

  filename = writeFile (TopLevel, True, NULL, NULL);

  if (filename) {
    FILE *f;

    if (!(f = fopen (filename, "wb")))
      perror ("fopen");
    else {
      unsigned char buf[174848];
      c64.c1541->SaveImageD64 (buf, 35);

      if (1 == fwrite (buf, sizeof buf, 1, f))
	status = True;
      else
	perror ("fwrite");
    }

    free (filename);
  }

  return status;
}

static Boolean do_hardReset (Boolean flag)
{
  if (!flag) return False;
  c64.Reset ();
  RasterCounter = 0;
  command = "";
  return True;
}

static Boolean do_softReset (Boolean flag)
{
  if (!flag) return False;
  c64.SoftReset ();
  command = "";
  return True;
}

static Boolean do_exit (Boolean flag)
{
  if (!flag) return False;
  XtCallActionProc (TopLevel, "Close", NULL, NULL, 0);
  return True;
}

static Boolean do_about (Boolean flag)
{
  if (!flag) return False;
  Dialog (TopLevel, "about", BT_OK);
  return True;
}

static Boolean toggle_emulate1541 (Boolean flag)
{
  if (!flag) return c64.C1541IsEnabled ();

  c64.EnableC1541 (flag = !appResources.emulate1541);
  appResources.emulate1541 = c64.C1541IsEnabled ();
  return flag == appResources.emulate1541;
}

static Boolean toggle_exrom (Boolean flag)
{
  if (!flag) return c64.CartridgeIsEnabled ();

  c64.EnableCartridge (flag = !appResources.exrom);
  appResources.exrom = c64.CartridgeIsEnabled ();
  return flag == appResources.exrom;
}
